#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon May 13 07:50:55 2019

@author: fuaa
"""
from matplotlib import pyplot as plt
import numpy as np
import getpass
import socket

_THRESHOLD = 10e-9


#########################################################################
def validar_resultado(*args, **kwargs):
    """"
    Función para validar resultado a invocar desde el notebook.
    """
    # _ES_CORRECTO = False
    _DEBUG = False

    # Abrir el cartel del validación.
    print("+-------------------------------------------------------------------------+")
    print("|               FuAA (1er. parcial 2022): validar resultado               |")
    print("+-------------------------------------------------------------------------+")
    for key, value in kwargs.items():
        if key == "debug":
            _DEBUG = value
    if _DEBUG:
        print('args:', args)
        print('kwargs:', kwargs)

    if (len(args) == 0):
        print("| Sin opciones para evaluar.                                              |")
        print("+-------------------------------------------------------------------------+")
        return False

    ###########################################################
    # Ejercicio 1. Regresión Lineal.
    ###########################################################
    elif (args[0] == "reg_lin"):
        if 'funcion' in kwargs:
            test_algoritmo = kwargs['funcion']
        else:
            print( "[validar_resultado] Error: llamada sin argumentos (%s)" % args[0] )
            return
        
        X = np.ones((20,2))
        X [:,1] = np.arange(20)
        y = (X[:,1] > 9)*2 -1
        w_ls_true = np.array([-1.42857143, 0.15037594 ])
       
        w_ls = test_algoritmo( X, y)
        
        condicion = son_iguales( w_ls_true , w_ls )
        fuaa_assert(    condicion, \
                        mensajeFalse = "El resultado no es válido.", \
                        mensajeTrue = "Resultado válido." )
        
    ###########################################################
    # Ejercicio 1. Transformación no lineal.
    ###########################################################
    elif (args[0] == "transformar_usando_polinomio_de_grado_g"):
        if 'funcion' in kwargs:
            test_algoritmo = kwargs['funcion']
        else:
            print("[validar_resultado] Error: llamada sin argumentos (%s)" % args[0])
            return

        pts = np.array([[1, 1, 1], [1, 2, 3], [1, 0, 2], [1, 2, 0]])
        pts_t_true = np.array([[1., 1., 1., 1., 1., 1., 1., 1., 1., 1., 1., 1., 1., 1., 1.],
                               [1., 2., 3., 4., 6., 9., 8., 12., 18., 27., 16., 24., 36., 54., 81.],
                               [1., 0., 2., 0., 0., 4., 0., 0., 0., 8., 0., 0., 0., 0., 16.],
                               [1., 2., 0., 4., 0., 0., 8., 0., 0., 0., 16., 0., 0., 0., 0.]])
        pts_t = test_algoritmo(pts, 4)

        # Validar dimensión de datos transformados.
        condicion1 = pts_t.shape == pts_t_true.shape
        fuaa_assert(condicion1,
                    mensajeFalse="La dimensión de los datos transformados no es correcta.",
                    mensajeTrue="Dimensión de salida: resultado validado.")

        # Validar transformación.
        condicion2 = son_iguales(pts_t_true, pts_t)
        fuaa_assert(condicion2,
                    mensajeFalse="La transformación implementada no es correcta.",
                    mensajeTrue="Transformación: resultado validado.")
    
    ###########################################################
    # Ejercicio 1. Cota de generalización VC.
    ###########################################################
    elif (args[0] == "cota_vc"):
        if 'funcion' in kwargs:
            test_algoritmo = kwargs['funcion']
        else:
            print( "[validar_resultado] Error: llamada sin argumentos (%s)" % args[0] )
            return
        
        cota_true = 2.410359620549163
       
        cota = test_algoritmo(159, 19, 0.01)
        
        condicion = son_iguales( cota_true , cota )
        fuaa_assert(    condicion, \
                        mensajeFalse = "El resultado no es válido.", \
                        mensajeTrue = "Resultado válido." )
     
    ###########################################################
    # Ejercicio 1. Regresión Ridge.
    ###########################################################
    elif (args[0] == "reg_ridge"):
        if 'funcion' in kwargs:
            test_algoritmo = kwargs['funcion']
        else:
            print( "[validar_resultado] Error: llamada sin argumentos (%s)" % args[0] )
            return
        
        X = np.ones((20,2))
        X [:,1] = np.arange(20)
        y = (X[:,1] > 9)*2 -1
        w_ls_true = np.array([-1.24183007,  0.13071895])
       
        w_ls = test_algoritmo( X, y, 100)
        
        condicion = son_iguales( w_ls_true , w_ls )
        fuaa_assert(    condicion, \
                        mensajeFalse = "El resultado no es válido.", \
                        mensajeTrue = "Resultado válido." )
        
    ###########################################################
    # Ejercicio 1. Error binario.
    ###########################################################
    elif (args[0] == "error_binario"):
        if 'funcion' in kwargs:
            test_algoritmo = kwargs['funcion']
        else:
            print( "[validar_resultado] Error: llamada sin argumentos (%s)" % args[0] )
            return

        y = np.array([1, 1, 1, 1, 1, -1, -1, -1, -1, -1])
        y_pred = np.array([-1, 1, 1, 1, 1, -1, -1, -1, -1, 1])
        err = test_algoritmo(y, y_pred)

        # Validar dimensión de la salida.
        condicion1 = not isinstance(err, np.ndarray)
        fuaa_assert(   condicion1, \
                    mensajeFalse = "La salida no tiene la dimensión correcta.", \
                    mensajeTrue= "Dimensión de salida: resultado validado.")

        # Validar transformación.
        condicion2 = err == 0.2
        fuaa_assert(   condicion2, \
                    mensajeFalse = "Error binario: resultado no validado.", 
                    mensajeTrue = "Error binario: resultado validado.")

    ###########################################################
    # Test.
    ###########################################################
    elif (args[0] == "test"):
        if validar_parametros(args, 4):
            condicion = args[1]
            mensajeF = args[2]
            mensajeT = args[3]
            fuaa_assert(condicion, mensajeFalse=mensajeF, mensajeTrue=mensajeT)

    ###########################################################
    # No hay ninguna opción de ejercicio.
    ###########################################################
    else:
        printcolor("Ninguna opción revisada.")

    # Cerrar el cartel.
    print("+-------------------------------------------------------------------------+")


#########################################################################
def identificar_parcial():
    username = getpass.getuser()
    hostname = socket.gethostname()
    print("Usuario %s en %s." % (username, hostname))


#########################################################################
def printcolor(mensaje, color="k"):
    """
    Función para imprimir en colores y con el formato de interés.
    """
    if (color == "r"):
        mensajeColor = "\x1b[31m" + mensaje + "\x1b[0m"
    elif (color == "g"):
        mensajeColor = "\x1b[32m" + mensaje + "\x1b[0m"
    elif (color == "y"):
        mensajeColor = "\x1b[33m" + mensaje + "\x1b[0m"
    elif (color == "b"):
        mensajeColor = "\x1b[34m" + mensaje + "\x1b[0m"
    elif (color == "p"):
        mensajeColor = "\x1b[35m" + mensaje + "\x1b[0m"
    elif (color == "c"):
        mensajeColor = "\x1b[36m" + mensaje + "\x1b[0m"
    else:
        mensajeColor = mensaje
    mensaje_out = " " + mensajeColor
    print(mensaje_out)


#########################################################################
def fuaa_assert(condicion,
                mensajeFalse="El resultado no es válido.",
                mensajeTrue="Resultado validado."):
    """
    Función similar al assert.
    """
    # Custom assert.
    if (condicion):
        printcolor(mensajeTrue, "g")
    else:
        printcolor(mensajeFalse, "r")


#########################################################################
def son_iguales(x1, x2, threshold=_THRESHOLD):
    """
    Evaluar si dos elementos son iguales o no, con una tolerancia dada (threshold).
    """
    if isinstance(x1, float) or isinstance(x1, int):
        dif = np.abs(x1 - x2)
        condicion = (dif < threshold)
    elif isinstance(x1, np.ndarray):
        dif = np.sqrt(np.sum((x1 - x2)**2)) / x1.size
        condicion = (dif < threshold)
    else:
        printcolor("Ningún tipo validado para son_iguales()", "r")
        condicion = False
    return condicion


#########################################################################
def validar_parametros(parametros, min_params):
    """
    Validar número de parámetros.
    """
    condicion = (len(parametros) >= min_params)
    if not condicion:
        printcolor("[validar_resultado] Insuficientes parámetros (\"%s\"), se necesitan %d, hay %d." %
                   (parametros[0], min_params, len(parametros)), "r")
    return condicion


#########################################################################
def visualizar_conjunto(X, y, titulo, dpi=100):
    '''
    Entrada: 
        X: matriz de tamaño Nx3 que contiene el conjunto de puntos de entrenamiento expresados en
           en coordenadas homogeneas. La primera coordenada de cada punto es uno.
        y: etiquetas asignadas a los puntos
    '''
    plt.figure(figsize=(8, 8))

    # Se grafican los puntos sorteados
    plt.scatter(X[y == 1, 1], X[y == 1, 2], s=40,
                color='b', marker='o', label='etiqueta 1')
    plt.scatter(X[y == -1, 1], X[y == -1, 2], s=40,
                color='r', marker='*', label='etiqueta -1')

    plt.legend()
    plt.axis('equal')
    plt.xlabel('x1')
    plt.ylabel('x2')
    plt.grid()
    plt.title(titulo)
    # plt.title('Conjunto de entrenamiento generado')


#########################################################################
def visualizar_modelo_lineal(X, y, w_g):
    '''
    Entrada: 
        X: matriz de tamaño Nx3 que contiene el conjunto de puntos de entrenamiento expresados en
           en coordenadas homogeneas. La primera coordenada de cada punto es uno.
        y: etiquetas asignadas a los puntos
        w_g: parámetros del modelo lineal encontrados 
    '''
    plt.figure(figsize=(8, 8))

    # Se grafican los puntos sorteados
    plt.scatter(X[y == -1, 1], X[y == -1, 2], s=40,
                color='r', marker='*', label='etiqueta -1')
    plt.scatter(X[y == 1, 1], X[y == 1, 2], s=40,
                color='b', marker='o', label='etiqueta 1')

    x1_min = X[:, 1].min()
    x1_max = X[:, 1].max()
    x1 = np.linspace(x1_min, x1_max)
    if w_g[2] == 0:
        x2_g = -(w_g[0]/w_g[2])*np.ones(x1.shape)
    else:
        # Se grafica la superficie de decisión encontrada
        x2_g = - w_g[1]/w_g[2] * x1 + -w_g[0]/w_g[2]
    plt.plot(x1, x2_g, label='funcion encontrada')

    plt.legend()
    plt.axis('equal')
    plt.xlabel('x1')
    plt.ylabel('x2')
    plt.title('Modelo lineal encontrado')


###############################################################################
def visualizar_costo_entrenamiento(costo_entrenamiento, xlabel='Número de iteración',
                                   ylabel='Costo de entrenamiento', title='Costo de entrenamiento'):
    '''
    Entrada:
        costo_entrenamiento: vector de dimención Niter con el costo en las Niter
                             iteraciones de entrenamiento de un modelo. 
    '''
    plt.figure(figsize=(8, 8))
    plt.plot(costo_entrenamiento, '*-')
    plt.xlabel(xlabel)
    plt.ylabel(ylabel)
    plt.title(title)
    plt.grid()


###############################################################################
def visualizar_frontera_decision(X, y, w, transformacion):
    '''
    Entrada:
        X: matriz de Nx3 que contiene los puntos en el espacio original
        y: etiquetas de los puntos
        w: vector de tamaño 10 que contiene los parámetros encontrados
    '''

    # Se construye una grilla de 50x50 en el dominio de los datos
    xs = np.linspace(X[:, 1].min(), X[:, 1].max())
    ys = np.linspace(X[:, 2].min(), X[:, 2].max())

    XX, YY = np.meshgrid(xs, ys)
    Z = np.zeros_like(XX)

    # se transforman los puntos de la grilla
    pts_grilla = np.vstack((np.ones(XX.size), XX.ravel(), YY.ravel())).T
    pts_grilla_transformados = transformacion(pts_grilla)

    # los puntos transformados son proyectados utilizando el w
    Z = pts_grilla_transformados @ w
    Z = Z.reshape(XX.shape)

    # se grafica la frontera de decisión, es decir, la línea de nivel 0
    plt.figure(figsize=(8, 8))
    plt.axis('equal')
    plt.contour(XX, YY, Z, [0])
    plt.scatter(X[:, 1][y == 1], X[:, 2][y == 1], s=40, color='b', marker='o',
                label='etiqueta -1')
    plt.scatter(X[:, 1][y == -1], X[:, 2][y == -1], s=40, color='r', marker='x',
                label='etiqueta 1')
    plt.title('Frontera de decision obtenida mediante\n%s()' %
              (transformacion.__name__))


###############################################################################
def generar_semianillos(N, radio1, radio2, ancho1, ancho2, delta_x, delta_y):
    '''
    Entrada:
        N: número de muestras a generar
        radio1: radio interior del semicírculo asociado a la clase 1
        radio2: radio interior del semicírculo asociado a la clase 2
        ancho1: diferencia entre el radio exterior e interior del semicírculo asociado a la clase 1
        ancho2: diferencia entre el radio exterior e interior del semicírculo asociado a la clase 2
        delta_x: corrimiento en x respecto al origen del semicirculo asociado a la clase 2
        delta_y: corrimiento en y respecto al origen del semicirculo asociado a la clase 2

    Salida:
        X: matriz de Nx3 que contiene los datos generados en coordenadas homogéneas
        y: etiquetas asociadas a los datos
    '''

    X = np.ones((N, 3))
    # se sortea a que clase pertenecen las muestras
    y = 2 * (np.random.rand(N) < 0.5) - 1

    # radios y ángulos del semicírculo superior
    radios = radio1*(y == 1) + radio2*(y == -1) + \
        (ancho1*(y == 1)+ancho2*(y == -1)) * np.random.rand(N)
    thetas = np.pi * np.random.rand(N)
    # coordenadas en x de ambos semicírculos
    X[:, 1] = radios * np.cos(thetas) * y + \
        (radio2 + ancho2/2 + delta_x)*(y == -1)
    # coordenadas en y de ambos semicírculos
    X[:, 2] = radios * np.sin(thetas) * y - delta_y * (y == -1)

    return X, y


###############################################################################
def generar_semianillos(N, radio1, radio2, ancho1, ancho2, delta_x, delta_y):
    '''
    Entrada:
        N: número de muestras a generar
        radio1: radio interior del semicírculo asociado a la clase 1
        radio2: radio interior del semicírculo asociado a la clase 2
        ancho1: diferencia entre el radio exterior e interior del semicírculo 
                asociado a la clase 1
        ancho2: diferencia entre el radio exterior e interior del semicírculo 
                asociado a la clase 2
        delta_x: corrimiento en x respecto al origen del semicirculo asociado a 
                 la clase 2
        delta_y: corrimiento en y respecto al origen del semicirculo asociado a 
                 la clase 2

    Salida:
        X: matriz de Nx3 que contiene los datos generados en coordenadas homogéneas
        y: etiquetas asociadas a los datos
    '''

    X = np.ones((N, 3))
    # se sortea a que clase pertenecen las muestras
    y = 2 * (np.random.rand(N) < 0.5) - 1

    # radios y ángulos del semicírculo superior
    radios = radio1 * (y == 1) + radio2 * (y == -1) + (
        ancho1 * (y == 1) + ancho2 * (y == -1)) * np.random.rand(N)
    thetas = np.pi * np.random.rand(N)
    # coordenadas en x de ambos semicírculos
    X[:, 1] = radios * np.cos(thetas) * y + (radio2 + ancho2 / 2 +
                                             delta_x) * (y == -1)
    # coordenadas en y de ambos semicírculos
    X[:, 2] = radios * np.sin(thetas) * y - delta_y * (y == -1)

    return X, y


###############################################################################
def generar_datos(N=100):
    r1 = 20
    w1 = 5
    r2 = 15
    w2 = -5
    off_x = -35
    off_y = -4
    X, y = generar_semianillos(N // 2, r1, r2, w1, w2, off_x, off_y)
    r1 = 20
    w1 = 5
    r2 = 15
    w2 = -5
    off_x = 10
    off_y = -7
    X_2, y_2 = generar_semianillos(N // 2, r1, r2, w1, w2, off_x, off_y)
    X = np.concatenate((X, X_2))
    y = np.concatenate((y, y_2))

    return X, y


###############################################################################
def visualizar_frontera_decision(X, y, w, grado, f_trans, titulo='Frontera de decision obtenida mediante transformación no lineal de datos'):
    '''
    Entrada:
        X: matriz de Nx3 que contiene los puntos en el espacio original
        y: etiquetas de los puntos
        w: vector de tamaño 10 que contiene los parámetros encontrados
    '''

    # Se construye una grilla de 50x50 en el dominio de los datos
    xs = np.linspace(X[:, 1].min() - 40, X[:, 1].max() + 40, num=100)
    ys = np.linspace(X[:, 2].min() - 40, X[:, 2].max() + 40, num=100)

    XX, YY = np.meshgrid(xs, ys)
    Z = np.zeros_like(XX)

    # se transforman los puntos de la grilla
    pts_grilla = np.vstack((np.ones(XX.size), XX.ravel(), YY.ravel())).T
    pts_grilla_transformados = f_trans(pts_grilla, grado)

    # los puntos transformados son proyectados utilizando el w
    Z = pts_grilla_transformados @ w
    Z = Z.reshape(XX.shape)

    # se grafica la frontera de decisión, es decir, la línea de nivel 0
    plt.figure(figsize=(8, 8))
    # plt.axis('equal')
    plt.pcolor(XX, YY, Z < 0, cmap='bwr', alpha=0.3)
    plt.contour(XX, YY, Z, [0])
    #bien = np.sign(np.dot(transformar_usando_polinomio_de_grado_g(X,12),w))==y
    #mal = np.sign(np.dot(transformar_usando_polinomio_de_grado_g(X,12),w))!=y
    plt.scatter(X[:, 1][y == 1],
                X[:, 2][y == 1],
                s=40,
                color='b',
                marker='o',
                label='etiqueta -1')
    plt.scatter(X[:, 1][y == -1],
                X[:, 2][y == -1],
                s=40,
                color='r',
                marker='*',
                label='etiqueta 1')
    plt.title(titulo)
    plt.xlim(-50, 50)
    plt.ylim(-50, 50)
    plt.legend()
